"use strict";

exports.__esModule = true;
exports.createTask = createTask;
exports.pushTask = pushTask;
exports.removeTask = removeTask;
exports.updateTask = updateTask;
exports.handleTasks = handleTasks;

var _lodash = _interopRequireDefault(require("lodash"));

var _guid = _interopRequireDefault(require("guid"));

var _chalk = _interopRequireDefault(require("chalk"));

var _util = _interopRequireDefault(require("util"));

var types = _interopRequireWildcard(require("../types"));

var workerActions = _interopRequireWildcard(require("./worker"));

var userActions = _interopRequireWildcard(require("./user"));

var globalActions = _interopRequireWildcard(require("./global"));

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

/**
 * Creates and pushes a task, this will trigger unoccupied workers to execute it.
 * The result will be a promise that resolves to the finished task
 *
 * @param {any | Array<any>} commands
 * @param {IdType | void} user
 * @param {IdType = guid.raw()} task
 * @param {IdType} worker
 */
function createTask(commands, user, task = _guid.default.raw(), worker) {
  return (dispatch, getState) => {
    const {
      users,
      workers
    } = getState();

    if (user !== undefined && !users[user]) {
      dispatch(globalActions.warn(`user doesn't exist!`, {
        user,
        worker,
        task
      }));
      return Promise.reject(`User ${user} doesn't exist`);
    }

    let callbacks = [],
        promise = new Promise((res, rej) => callbacks = [res, rej]);
    dispatch(globalActions.info(`accepted ${_util.default.inspect(commands, {
      depth: null
    }).slice(0, 256)}`, {
      owner: 'task',
      user,
      task,
      worker
    }));
    commands = Array.isArray(commands) ? commands : [commands];
    if (user) dispatch(userActions.updateUser(user));
    dispatch(pushTask(task, user, commands, promise, ...callbacks, worker));
    return promise;
  };
}
/**
 * Pushes a global or worker-bound task
 *
 * @param {IdType} task
 * @param {IdType | void} user
 * @param {Array<string>} commands
 * @param {Promise<any>} promise
 * @param {any} resolve
 * @param {any} reject
 * @param {string | void} worker
 * @param {number} timestamp
 */


function pushTask(task, user, commands, promise, resolve, reject, worker, timestamp) {
  return {
    type: types.pushTask,
    task,
    user,
    commands,
    promise,
    resolve,
    reject,
    worker,
    timestamp
  };
}
/**
 * Removes a task, this is used by the worker once it has resolved a task
 *
 * @param {IdType} task
 * @param {IdType} worker
 */


function removeTask(task, worker) {
  return {
    type: types.removeTask,
    worker,
    task
  };
}
/**
 * Update task
 *
 * @param {IdType} task
 * @param {IdType | void} worker
 * @param {TaskNullableType} props
 */


function updateTask(task, worker, props) {
  return _extends({
    type: types.updateTask,
    worker,
    task
  }, props);
}
/**
 * Observe tasks and workers, every reaction triggers the observer to look for unfulfileld
 * tasks that should then be applied to workers signaling their availability by the occupid flag
 */


function handleTasks() {
  return async (dispatch, getState) => {
    try {
      let state = getState();

      if (state.tasks.length) {
        // Fetch all tasks belonging the first user in the queue
        const first = state.tasks[0]; // Search for worker candidate, start by picking the first available

        let worker;

        for (let candidate of Object.keys(state.workers).map(key => state.workers[key])) {
          // Worker should be alive
          if (!candidate.alive) continue;else if (candidate.lastUser === first.user) {
            // We prefer workers that worked for the same user previously
            worker = candidate;
            break;
          } else if (candidate.occupied) // It shouldn't be occupied
            continue;else if (!worker || worker.timestamp < candidate.timestamp) // Otherwise we pick a worker with the oldest timestamp
            worker = candidate;
        } // If a worker was found, on to providing it tasks to execute ...


        if (worker) {
          const {
            id,
            socket
          } = worker; // Claim worker and remove group from the global state object, this will cause another
          // observe-ping so other users waiting in line can be processed as well ...
          // claimTasks returns a resolver that needs to be called once all actions are through.

          let tasks = state.tasks.filter(task => task.user === first.user); // Add tasks to occupied worker

          if (worker.occupied) {
            dispatch(globalActions.info(`adds ${tasks.length} tasks`, {
              owner: 'worker',
              worker: id,
              user: first.user
            }));
            return dispatch({
              type: types.claimWorker,
              worker: id,
              user: first.user,
              tasks
            });
          } // Claim worker, add tasks to its own collection


          dispatch(workerActions.claimWorker(id, first.user, tasks));

          if (worker.lastUser !== first.user) {
            // Store the last users state
            if (worker.lastUser && getState().users[worker.lastUser]) await dispatch(userActions.storeState(worker.lastUser, id)); // Clear worker

            await dispatch(workerActions.clear(id)); // Load users last state, if present

            await dispatch(userActions.loadState(first.user, id));
          } // Execute all tasks, this must be a loop because tasks can be added at any time ...


          while (getState().workers[id]) {
            // Tasks should be sorted after their timestamp, some could have been interrupted by worker fallouts
            const tasks = _lodash.default.sortBy(getState().workers[id].tasks, ['timestamp']);

            if (tasks && tasks.length > 0) for (let task of tasks) await dispatch(workerActions.execute(id, task));else break;
          }

          dispatch(workerActions.releaseWorker(id));
        }
      }
    } catch (error) {
      dispatch(globalActions.error(error));
    }
  };
}