"use strict";

exports.__esModule = true;
exports.emit = emit;
exports.emitFrame = emitFrame;
exports.addUser = addUser;
exports.removeUser = removeUser;
exports.updateUser = updateUser;
exports.storeState = storeState;
exports.loadState = loadState;
exports.clearStates = clearStates;
exports.undoRedo = undoRedo;
exports.handleUser = handleUser;

var _lodash = _interopRequireDefault(require("lodash"));

var _path = _interopRequireDefault(require("path"));

var _chalk = _interopRequireDefault(require("chalk"));

var _fsExtra = _interopRequireDefault(require("fs-extra"));

var types = _interopRequireWildcard(require("../types"));

var workerActions = _interopRequireWildcard(require("./worker"));

var taskActions = _interopRequireWildcard(require("./task"));

var globalActions = _interopRequireWildcard(require("./global"));

var _defaults = _interopRequireDefault(require("../../defaults"));

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Sends data via SocketIO wrapped around a Promise which resolves once the client acknowledges
 *
 * @param {*} socket
 * @param {*} data
 */
const emitAsync = (socket, data) => new Promise((resolve, reject) => {
  try {
    socket.emit(data instanceof Buffer ? 'binary' : 'json', data, resolve);
  } catch (error) {
    reject(error);
  }
});
/**
 * Emits data back to the user via socket or collects it internally under task.results. Emit is queued.
 *
 * @param {IdType} user
 * @param {Object | Array<Object>} data
 * @param {IdType} task
 * @param {IdType | void} worker
 * @param {boolean = false} wrap - Wrap the message into Begin- & EndFrame pairs
 */


function emit(user, data, task, worker, wrap = false) {
  return (dispatch, getState) => {
    data = Array.isArray(data) ? data : [data];
    const {
      users,
      config
    } = getState();
    const userObj = users[user];

    if (userObj && userObj.socket) {
      // If the user is bound to a socket, respond there ...
      const queue = userObj.queue.then(async () => {
        // 1. Send BeginFrame to signal message start
        if (wrap) await emitAsync(userObj.socket, {
          command: 'Result',
          from: 'BeginFrame',
          result: '',
          transactionID: task
        }); // 2. Send contents

        await Promise.all(data.map(item => emitAsync(userObj.socket, item))); // 3. Send EndFrame to signal end of message

        if (wrap) emitAsync(userObj.socket, {
          command: 'Result',
          from: 'EndFrame',
          result: '',
          transactionID: task
        });
      }).catch(error => dispatch(globalActions.error(error, {
        owner: 'user',
        user,
        task
      }))); // Update the users internal promise-queue

      dispatch(updateUser(user, {
        queue
      }));
      return queue;
    } else {
      data.forEach(item => {
        // If the user doesn't have a socket it should collect results in the task given
        if (!userObj || config.collectResultsOnly) {
          // Collect data only, this is for internal commands (user=undefined) or programmatic tasks
          item.result && dispatch(taskActions.updateTask(task, worker, {
            results: [item.result],
            streamData: item.streamData
          }));
        } else {
          // Otherwise everything is collected, this is for rest-endpoints for instance ...
          dispatch(taskActions.updateTask(task, worker, {
            results: [item]
          }));
        }
      });
    }
  };
}
/**
 * Emits Begin- & End-framed data back to the user via socket or collects it internally under task.results. Emit is queued.
 *
 * @param {IdType} user
 * @param {Object | Array<Object>} data
 * @param {IdType} task
 * @param {IdType | void} worker
 */


function emitFrame(user, data, task, worker) {
  return emit(user, data, task, worker, true);
}
/**
 * Adds new user
 *
 * @param {IdType} user
 * @param {*} socket
 */


function addUser(user, socket) {
  return (dispatch, getState) => {
    //dispatch(globalActions.info(`accepted ${user}`, { owner: 'user', user }))
    const {
      output
    } = getState().config;

    const folder = _path.default.join(output, 'states', `${user}`);

    if (_fsExtra.default.existsSync(folder)) _fsExtra.default.removeSync(folder);

    _fsExtra.default.mkdirSync(folder);

    dispatch({
      type: types.addUser,
      user,
      socket,
      folder
    });
  };
}
/**
 * Removes existing user
 *
 * @param {IdType} user
 */


function removeUser(user) {
  return (dispatch, getState) => {
    const {
      users,
      config
    } = getState();
    const userObj = users[user];

    if (userObj) {
      dispatch(globalActions.info(`terminating ${user}`, {
        owner: 'user',
        user
      }));
      dispatch({
        type: types.removeUser,
        user
      });
      setTimeout(() => {
        try {
          if (_fsExtra.default.existsSync(userObj.folder)) _fsExtra.default.removeSync(userObj.folder);
        } catch (error) {
          dispatch(globalActions.error(`remove state folder throws an error: ${error}`, {
            owner: 'user',
            user
          }));
        }
      }, config.stateDebounce + 5000);
    } // Remove any trace or leftover the user has left


    getState().collections.workers.forEach(w => getState().workers[w].lastUser === user && dispatch(workerActions.updateWorker(w, {
      lastUser: undefined
    })));
  };
}
/**
 * Updates existing user
 *
 * @param {IdType} user
 * @param {UserNullableType} updates
 */


function updateUser(user, updates) {
  return {
    type: types.updateUser,
    user,
    updates
  };
}
/**
 * Stores user state in its local state-folder
 *
 * @param {IdType} user
 * @param {IdType} worker
 */


function storeState(user, worker, task) {
  return async (dispatch, getState) => {
    const {
      users,
      config
    } = getState();
    const userObj = users[user]; // Only store if actions have mutated the users state

    if (userObj && userObj.actions > 0) {
      // Delete newer states, this means the user undid something and started to work on that state
      try {
        if (userObj.systemFile) userObj.stateFiles.filter(state => parseInt(state) > parseInt(userObj.systemFile)).forEach(state => _fsExtra.default.unlinkSync(_path.default.join(userObj.folder, state)));
      } catch (error) {
        dispatch(globalActions.error(error, {
          owner: 'user',
          user,
          worker,
          task
        }));
      } // Prepare a new state, then let classcad write it


      let response,
          systemFile = `${_defaults.default.now()}.state`,
          command = {
        command: 'StoreState',
        path: _path.default.join(userObj.folder, systemFile)
      };
      if (worker) response = await dispatch(workerActions.execute(worker, [command]));else response = await dispatch(taskActions.createTask(command, user, task)); // List and sort all states

      const stateFiles = _fsExtra.default.readdirSync(userObj.folder).sort((a, b) => parseInt(a) > parseInt(b)); // Remove trailing state, if present


      try {
        if (stateFiles.length > config.maxStates) _fsExtra.default.unlinkSync(_path.default.join(userObj.folder, stateFiles.shift()));
      } catch (error) {
        dispatch(globalActions.error(error, {
          owner: 'user',
          user,
          worker,
          task
        }));
      } // Update system-state file and reset actions to zero


      dispatch(updateUser(user, {
        systemFile,
        stateFiles,
        actions: 0
      })); // Notify user since states have changed

      userObj.socket && userObj.socket.emit('undo', {
        count: stateFiles.length,
        current: stateFiles.length - 1
      });
      return response;
    }
  };
}
/**
 * loads the most recent user state from its local user state-folder.
 *
 * @param {IdType} user
 * @param {IdType} worker
 */


function loadState(user, worker, task, name) {
  return async (dispatch, getState) => {
    const userObj = getState().users[user];

    if (userObj) {
      const {
        systemFile,
        folder,
        stateFiles,
        socket
      } = userObj;

      const statePath = (name || systemFile) && _path.default.join(folder, name || systemFile);

      if (_fsExtra.default.existsSync(statePath)) {
        let response,
            command = {
          command: 'LoadState',
          path: statePath,
          recalc: false,
          updateGraphics: !!name,
          isUndoOrRedo: !!name
        };
        if (worker) response = await dispatch(workerActions.execute(worker, [command]));else response = await dispatch(taskActions.createTask(command, user, task));

        if (name) {
          // When an explicit name was given (undo/redo), update systemFile and notify user
          dispatch(updateUser(user, {
            systemFile: name
          }));
          socket && socket.emit('undo', {
            count: stateFiles.length,
            current: stateFiles.indexOf(name)
          });
        }

        return response;
      }
    }
  };
}
/**
 * Clear all states.
 *
 * @param {IdType} user
 */


function clearStates(user) {
  return async (dispatch, getState) => {
    const userObj = getState().users[user];

    if (userObj) {
      try {
        userObj.socket && userObj.socket.emit('undo', {
          count: 0,
          current: 0
        });
        userObj.stateFiles.forEach(state => userObj.systemFile !== state && _fsExtra.default.unlinkSync(_path.default.join(userObj.folder, state)));
      } catch (error) {
        //dispatch(globalActions.error(error, { owner: 'user', user }))
        dispatch(globalActions.error(`${error.stack} ${JSON.stringify(userObj)}`, {
          owner: 'user',
          user
        }));
      } finally {
        dispatch(updateUser(user, {
          stateFiles: []
        }));
      }
    }
  };
}
/**
 * Either call loadState or cause a simple ping in order to prevent choking
 * the client which might be waiting for a response.
 *
 * @param {IdType} user
 * @param {'undo' | 'redo'} type
 */


function undoRedo(user, type, task) {
  return async (dispatch, getState) => {
    const userObj = getState().users[user];

    if (userObj) {
      const {
        stateFiles,
        systemFile,
        id
      } = userObj;
      const state = stateFiles[stateFiles.indexOf(systemFile) + (type === 'undo' ? -1 : 1)];
      if (state) dispatch(loadState(id, undefined, task, state));else dispatch(taskActions.createTask({}, id, task));
    }
  };
}
/**
 * Deals with a users socket and dispatches actions
 *
 * @param {*} socket
 */


function handleUser(socket) {
  return async (dispatch, getState) => {
    let debouncedStoreState = _lodash.default.debounce(id => getState().globals.running && getState().users[id] && dispatch(storeState(id)), getState().config.stateDebounce);

    let id;
    socket.on('error', error => dispatch(globalActions.error(error)));
    socket.on('disconnect', () => dispatch(removeUser(id)));
    socket.once('init', options => {
      id = socket.id;
      dispatch(addUser(id, socket));
      socket.emit('permission', {
        id,
        version: '8.x.x'
      });
      socket.on('send', async (contextId, command) => {
        dispatch(taskActions.createTask(command, id, contextId));
        debouncedStoreState(id);
      });
      socket.on('undo', task => dispatch(undoRedo(id, 'undo', task)));
      socket.on('redo', task => dispatch(undoRedo(id, 'redo', task)));
    });
  };
}