"use strict";

exports.__esModule = true;
exports.addWorker = addWorker;
exports.updateWorker = updateWorker;
exports.removeWorker = removeWorker;
exports.execute = execute;
exports.clear = clear;
exports.resolveWorkerTask = resolveWorkerTask;
exports.claimWorker = claimWorker;
exports.releaseWorker = releaseWorker;
exports.emitWorkerTask = emitWorkerTask;
exports.handleWorker = handleWorker;
exports.spawnWorker = spawnWorker;

var _lodash = _interopRequireDefault(require("lodash"));

var _path = _interopRequireDefault(require("path"));

var _fsExtra = _interopRequireDefault(require("fs-extra"));

var _guid = _interopRequireDefault(require("guid"));

var _chalk = _interopRequireDefault(require("chalk"));

var _util = _interopRequireDefault(require("util"));

var _delay = _interopRequireDefault(require("delay"));

var _child_process = require("child_process");

var _redux = require("redux");

var types = _interopRequireWildcard(require("../types"));

var userActions = _interopRequireWildcard(require("./user"));

var taskActions = _interopRequireWildcard(require("./task"));

var globalActions = _interopRequireWildcard(require("./global"));

var _helpers = require("../../helpers");

var _defaults = _interopRequireDefault(require("../../defaults"));

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Adds new worker
 *
 * @param {IdType} worker
 * @param {*} socket
 */
function addWorker(worker, socket) {
  return dispatch => {
    dispatch(globalActions.info(`accepted`, {
      owner: 'worker',
      worker
    }));
    dispatch({
      type: types.addWorker,
      worker,
      socket
    });
  };
}
/**
 * Updates existig worker
 *
 * @param {IdType} worker
 * @param {WorkerNullableType} updates
 */


function updateWorker(worker, updates) {
  return {
    type: types.updateWorker,
    worker,
    updates
  };
}
/**
 * Removes existing worker. Force==true will kill the worker right away, then collect open tasks in order
 * to push them to the global queue again so another unoccupied worker may pick them up. If force==false
 * then it will wait until the worker has resolved its open tasks first, then kill it.
 *
 * @param {IdType} worker
 * @param {boolean = false} force
 */


function removeWorker(worker, force = false) {
  return async (dispatch, getState) => {
    const s = getState();
    const w = s.workers[worker];

    if (w) {
      dispatch(updateWorker(worker, {
        alive: false
      }));

      if (force) {
        if (s.config.enableTaskSalvaging) {
          // Transfer remaining tasks over to the global collection so they can be processed again
          w.tasks.forEach(t => {
            if (t.user) {
              const args = {
                owner: 'worker',
                worker,
                user: t.user || w.lastUser,
                task: t.id
              };
              dispatch(globalActions.warn(`salvaging ${_util.default.inspect(t.commands, {
                depth: null
              }).replace(/\s+/g, ' ')}`, args));
              dispatch(taskActions.pushTask(t.id, t.user, t.commands, t.promise, t.resolve, t.reject, undefined, t.timestamp));
            }
          });
        }

        w.resolve && w.resolve();
      } else await w.sequence;

      dispatch({
        type: types.removeWorker,
        worker
      });
      w.process && w.process.kill();
      w.socket && w.socket.close();
      dispatch(globalActions.info(`terminated ${worker} ${force ? 'hard' : 'soft'}`, {
        owner: 'worker',
        worker
      }));
    }
  };
}
/**
 * Sends a task or a series of commands to a worker
 *
 * @param {IdType} worker
 * @param {*} payload
 */


function execute(worker, payload) {
  return (dispatch, getState) => {
    const {
      workers,
      users
    } = getState();
    const workerObj = workers[worker];

    if (workerObj) {
      const isTask = !Array.isArray(payload);
      const transactionID = isTask ? payload.id : _guid.default.raw();
      const user = isTask ? payload.user : undefined;
      const promise = isTask ? payload.promise : dispatch(taskActions.createTask(payload, user, transactionID, worker));
      const commands = isTask ? payload.commands : payload; // Count user actions

      if (user && users[user]) dispatch(userActions.updateUser(user, {
        actions: users[user].actions + 1
      })); // Wrap payload into BeginFrame/EndFrame pairs and send to worker

      workerObj.socket.send(JSON.stringify({
        command: 'BeginFrame',
        transactionID
      }));
      commands.forEach(command => {
        let slice = _util.default.inspect(command, {
          depth: null
        });

        if (slice.length > 256) slice = slice.slice(0, 256) + ' ...';
        dispatch(globalActions.log(`  sends ${slice.replace(/\s+/g, ' ')}`, {
          owner: 'worker',
          worker,
          user,
          task: transactionID
        }));
        workerObj.socket.send(JSON.stringify(command));
      });
      workerObj.socket.send(JSON.stringify({
        command: 'EndFrame',
        transactionID
      }));
      return promise;
    }

    return Promise.reject('worker not found');
  };
}
/**
 * Clears worker
 *
 * @param {IdType} worker
 */


function clear(worker) {
  return async (dispatch, getState) => {
    try {
      if (getState().workers[worker]) {
        dispatch(globalActions.info(`clearing`, {
          owner: 'worker',
          worker
        }));
        await dispatch(execute(worker, [{
          command: 'Clear'
        }]));
      }
    } catch (error) {
      dispatch(globalActions.error(error, {
        owner: 'worker',
        worker
      }));
    }
  };
}
/**
 * This marks a workers task as completed and resolves & removes it
 *
 * @param {IdType} worker
 * @param {IdType} taskId
 */


function resolveWorkerTask(worker, taskId) {
  return (dispatch, getState) => {
    const task = getState().workers[worker].tasks.find(t => t.id === taskId);
    dispatch(taskActions.removeTask(taskId, worker));

    if (task) {
      task.duration = (_defaults.default.now() - task.timestamp) / 1000;
      dispatch(globalActions.info(`resolved`, {
        owner: 'task',
        user: task.user,
        task: taskId,
        worker,
        duration: task.duration
      }));
      task.resolve && task.resolve(task);
    }
  };
}
/**
 * Marks a worker as occupied, removes a group of tasks and returns a resolver that should be called
 * once all of the tasks have been resolved.
 *
 * @param {IdType} worker
 * @param {IdType} user
 * @param {Array<TaskType>} tasks
 */


function claimWorker(worker, user, tasks) {
  return dispatch => {
    dispatch(globalActions.info(`claims ${tasks.length} tasks`, {
      owner: 'worker',
      worker,
      user
    }));
    let resolve,
        sequence = new Promise(r => resolve = r);
    dispatch({
      type: types.claimWorker,
      worker,
      user,
      tasks,
      sequence,
      resolve
    });
  };
}
/**
 * Resolves worker and signals it free for the taking
 *
 * @param {IdType} worker
 */


function releaseWorker(worker) {
  return (dispatch, getState) => {
    const workerObj = getState().workers[worker];

    if (workerObj) {
      dispatch(globalActions.info(`released`, {
        owner: 'worker',
        worker
      }));
      workerObj.resolve && workerObj.resolve();
      dispatch({
        type: types.releaseWorker,
        worker
      });
    }
  };
}
/**
 * Deals with socket errors
 *
 * @param {IdType} worker
 * @param {Error} workerError
 */


function handleWorkerError(worker, workerError) {
  return (dispatch, getState) => {
    const running = getState().globals.running;
    if (running) dispatch(globalActions.error(workerError || new Error(`worker closed while running`), {
      owner: 'worker',
      worker
    }));
    dispatch(removeWorker(worker, running));
  };
}
/**
 * Forwards results to receiver, may be a users socket or the tasks result collection
 *
 * @param {IdType} worker
 * @param {IdType} task
 * @param {Object} data
 */


function emitWorkerTask(worker, task, data) {
  return (dispatch, getState) => {
    const {
      users,
      workers,
      config
    } = getState();
    const isBinary = data instanceof Buffer;
    const t = workers[worker].tasks.find(t => t.id === task);

    if (t) {
      const args = {
        owner: 'worker',
        worker,
        user: t.user,
        task,
        duration: (_defaults.default.now() - t.timestamp) / 1000
      };

      if (!isBinary && data.command === 'ErrorMessage') {
        let errorState = parseInt(data.attributes.errorState);

        switch (errorState //0: no error; 1: warning, 2: error go on; 3: error stop
        ) {
          case 0:
            dispatch(globalActions.info(data.attributes.errorMessage, args));
            break;

          case 1:
            dispatch(globalActions.warn(data.attributes.errorMessage, args));
            break;

          case 2:
          case 3:
          default:
            dispatch(globalActions.error(data.attributes.errorMessage, args));
            break;
        }
      }

      if (isBinary) {
        dispatch(globalActions.log(`received [binary]`, args));
      } else if (data.result) {
        let slice = _util.default.inspect(data.result, {
          depth: null
        });

        if (slice.length > 256) slice = slice.slice(0, 256) + ' ...';
        dispatch(globalActions.log(`received ${slice.replace(/\s+/g, ' ')}`, args));
      }

      dispatch(userActions.emit(t.user, data, task, worker));
    }
  };
}
/**
 * Deals with a workers socket and dispatches further actions
 *
 * @param {*} socket
 */


function handleWorker(socket) {
  return (dispatch, getState) => {
    let id;
    let task;
    let user;

    try {
      socket.on('error', error => dispatch(handleWorkerError(id, error)));
      socket.on('close', () => dispatch(handleWorkerError(id)));
      socket.on('message', data => {
        try {
          const isBinary = data instanceof Buffer;
          const {
            users,
            workers,
            globals
          } = getState();

          if (!isBinary) {
            // Convert data to JSON
            data = JSON.parse(data); // ClassCAD sends first handshake, which feeds back the id and then the ready state

            if (data.response === 'Handshake') {
              id = data.id;
            } else if (data.response === 'Ready') {
              dispatch(addWorker(id, socket));
            } else if (data.from === 'BeginFrame') {
              // BeginFrame marks the beginning of a sequence of commands and carries a task id
              task = workers[id].tasks.find(task => task.id === data.transactionID);
              user = task && users[task.id];
            }
          }

          if (task) {
            // Forward results to the users socket or the tasks result collection
            dispatch(emitWorkerTask(id, task.id, data));

            if (!isBinary && data.from === 'EndFrame') {
              // EndFrame makrs the end of a sequence of commands, resolves a task
              dispatch(resolveWorkerTask(id, task.id));
            }
          }
        } catch (error) {
          dispatch(globalActions.error(error, {
            owner: 'worker',
            worker: id,
            task: task ? task.id : undefined,
            user: user ? user.id : undefined
          }));
          if (task) task.reject(error);
        }
      }); // Allow ClassCAD to participate

      const {
        configurationData,
        compression,
        multipart,
        sendTreePatches,
        disableGraphics,
        enableInvisibleGraphics
      } = getState().config;
      socket.send(JSON.stringify({
        command: 'configuration',
        configurationData,
        compression,
        multipart,
        sendTreePatches,
        disableGraphics,
        enableInvisibleGraphics
      }));
    } catch (error) {
      dispatch(globalActions.error(error, {
        owner: 'worker',
        worker: id
      }));
    }
  };
}
/**
 * Spawns a local worker instance and awaits its ready state
 *
 * @param {number = 3} tries
 */


function spawnWorker(tries = 3) {
  return async (dispatch, getState, server) => {
    const id = _guid.default.raw();

    try {
      const spawned = getState().globals.spawned;
      dispatch(globalActions.updateGlobals({
        spawned: spawned + 1
      }));
      let {
        exec,
        execArgs,
        stdout,
        workerTimeout,
        showConsole
      } = getState().config;
      execArgs = [...execArgs, `--id`, `${id}`];
      if (spawned) dispatch(globalActions.info(`spawning ${id}`, {
        owner: 'worker',
        worker: id
      }));else dispatch(globalActions.info(`spawning ${id}\n${exec} ${execArgs.join(' ')}`, {
        owner: 'worker',
        worker: id
      })); // Execute process

      let ccProcess = (0, _child_process.spawn)(exec, execArgs, {
        stdio: stdout ? 'pipe' : undefined,
        shell: showConsole,
        detached: showConsole
      });

      if (stdout) {
        ccProcess.stdout.setEncoding('utf8').on('data', data => dispatch(globalActions.log(data.trim(), {
          owner: 'worker',
          worker: id
        })));
        ccProcess.stderr.setEncoding('utf8').on('data', data => dispatch(globalActions.error(data.trim(), {
          owner: 'worker',
          worker: id
        })));
      }

      ccProcess.on('error', error => dispatch(globalActions.error(error, {
        owner: 'worker',
        worker: id
      }))); // Wait for the worker to come up, then link it to the process we have started here

      const worker = await (0, _helpers.awaitState)(server.store, s => s.collections.workers, workers => workers.find(w => w === id), workerTimeout);

      if (worker) {
        dispatch(updateWorker(worker, {
          process: ccProcess
        })); // Making sure it respawns if something's wrong

        const unsub = (0, _helpers.subscribe)(server.store, state => state.workers[id] && !state.workers[id].alive, () => unsub() || getState().globals.running && dispatch(spawnWorker()));
        return id;
      } else {
        dispatch(globalActions.error(`worker died coming up`, {
          owner: 'worker',
          worker: id
        }));
        ccProcess && ccProcess.kill();
        ccProcess = undefined; // Try again, with a bit of delay

        if (tries > 0 && getState().globals.running) {
          return (await (0, _delay.default)(2000)) || getState().globals.running && dispatch(spawnWorker(tries - 1));
        }
      }
    } catch (error) {
      dispatch(globalActions.error(error, {
        owner: 'worker',
        worker: id
      }));
    }
  };
}